import argparse
import warnings
from pathlib import Path

import pandas as pd
from asreview import ASReviewData
from asreview.data.base import load_data


def _check_order_arg(order):
    # if no hierarchy is specified, set to default: "riu"
    if order is None:
        return "riu"

    allowed_orders = ["riu", "rui", "uri", "uir", "iru", "iur"]
    if order in allowed_orders:
        return order
    else:
        raise ValueError(
            f"hierarchy '{order}' not found, should be one of the"
            f" following: {allowed_orders}"
        )


def _check_resolve_arg(resolve):
    # if no resolve method is specified, set to default: "keep_one"
    if resolve is None:
        return "keep_one"

    allowed_resolve = ["keep_one", "keep_all", "abort"]
    if resolve in allowed_resolve:
        return resolve
    else:
        raise ValueError(
            f"conflict_resolve '{resolve}' not found, should be one "
            f"of the following: {allowed_resolve}"
        )


def _check_suffix(input_files, output_file):
    # Also raises ValueError on URLs that do not end with a file extension
    suffixes = [Path(item).suffix for item in input_files if item is not None]
    suffixes.append(Path(output_file).suffix)

    set_ris = {".txt", ".ris"}
    set_tabular = {".csv", ".tab", ".tsv", ".xlsx"}
    set_suffixes = set(suffixes)

    if len(set(suffixes)) > 1:
        if not (set_suffixes.issubset(set_ris) or set_suffixes.issubset(set_tabular)):
            raise ValueError(
                "Files with different file types were; all input files, as well as the"
                " output file, should be of the same type. "
            )


def _check_label_errors(as_lab, path_lab):
    if as_lab is not None:
        if as_lab.labels is None:
            warnings.warn(
                f"'{path_lab}' was passed as a labeled dataset but no labels were"
                " found, continuing with its records marked as unlabeled. If this is"
                " not correct, check if your data format complies with:"
                " https://asreview.readthedocs.io/en/latest/data_format.html",
                stacklevel=1,
            )


def _append_df(list_df, as_obj, label):
    # retrieve part of dataframe with label -1, 0 or 1
    df_slice = as_obj.df[as_obj.labels == label].reset_index(drop=True)

    if not df_slice.empty:
        list_df.append(df_slice)


def _concat_label(list_df, label, pid="doi"):
    # if there are any dataframes with the given label, concatenate and drop
    # duplicates on pid and title/abstract
    if list_df:
        df_all = pd.concat(list_df).reset_index(drop=True)
        df_all["included"] = label
        n_total = len(df_all)

        df_all = ASReviewData(df=df_all).drop_duplicates(pid=pid).reset_index(drop=True)

        n_total_dedup = n_total - len(df_all)
        print(
            f"Detected {n_total} records with label '{label}', from which"
            f" {n_total_dedup} duplicate records with the same label were removed."
        )
    else:
        df_all = pd.DataFrame()

    return df_all


def create_composition(
    rel_path=None,
    irr_path=None,
    lab_path=None,
    unl_path=None,
    pid="doi",
    order="riu",
    resolve="keep_one",
):
    # load all input files and URLs into ASReviewData objects, fill with None
    # if input was not specified
    input_files = [rel_path, irr_path, lab_path, unl_path]
    as_rel, as_irr, as_lab, as_unl = (
        load_data(item) if item is not None else None for item in input_files
    )

    # check whether input files are correctly labeled
    _check_label_errors(as_lab, lab_path)

    # create lists to append dataframes with a specific label to
    list_df_rel, list_df_irr, list_df_unl = [], [], []

    # split labeled input data in relevant, irrelevant and unlabeled and add
    # to list of dataframes for that label
    if as_lab is not None:
        if as_lab.labels is not None:
            _append_df(list_df_rel, as_lab, 1)
            _append_df(list_df_irr, as_lab, 0)
            _append_df(list_df_unl, as_lab, -1)
        else:
            list_df_unl.append(as_lab.df)

    # add dataframe to list of dataframes for that label
    if as_rel is not None:
        list_df_rel.append(as_rel.df)
    if as_irr is not None:
        list_df_irr.append(as_irr.df)
    if as_unl is not None:
        list_df_unl.append(as_unl.df)

    # concatenate all dataframes with the same label, drop duplicates and map
    # them in a dictionary
    dict_dfs = {
        "r": _concat_label(list_df_rel, 1, pid),
        "i": _concat_label(list_df_irr, 0, pid),
        "u": _concat_label(list_df_unl, -1, pid),
    }

    # map letters to corresponding term
    dict_terms = {"r": "relevant", "i": "irrelevant", "u": "unlabeled"}

    # concatenate in specified order, only the first duplicate entry is kept
    as_conflict = ASReviewData(
        df=pd.concat(
            [dict_dfs[order[0]], dict_dfs[order[1]], dict_dfs[order[2]]]
        ).reset_index(drop=True)
    )

    # check for label conflicts
    df_conflicting_dups = as_conflict.df[as_conflict.duplicated(pid)]
    if len(df_conflicting_dups) > 0:
        as_conflicts_only = ASReviewData(df=df_conflicting_dups.reset_index(drop=True))
        # create a dataframe with the relevant info for the user
        if pid in as_conflicts_only.df.columns:
            df_info_conflicts = pd.DataFrame(
                {
                    pid: as_conflicts_only.df[pid].fillna(""),
                    "Title": as_conflicts_only.title,
                    "Abstract": as_conflicts_only.abstract,
                }
            )
        else:
            df_info_conflicts = pd.DataFrame(
                {
                    "Title": as_conflicts_only.title,
                    "Abstract": as_conflicts_only.abstract,
                }
            )

        # pandas settings to print properly
        with pd.option_context(
            "display.max_rows",
            None,
            "display.max_columns",
            3,
            "max_colwidth",
            40,
            "display.width",
            500,
            "display.colheader_justify",
            "left",
        ):
            print(
                f"\nSome records have inconsistent labels in the input files. This may"
                " be intentional because you are trying to overwrite labels in an input"
                " file with labels from another input file. However, it may also be"
                " because some records are unintentionally labeled inconsistently.\n\n"
                "The following records have inconsistent labels in the input files:\n"
                f"{df_info_conflicts}\n"
            )

        if resolve == "abort":
            raise ValueError("Abort composing because inconsistent labels were found.")

        elif resolve == "keep_one":
            warnings.warn(
                f"Continuing, keeping one label for records with inconsistent labels,"
                " resolving conflicts using the following hierarchy:"
                f"\n1. {dict_terms[order[0]]}\n2. {dict_terms[order[1]]}"
                f"\n3. {dict_terms[order[2]]}",
                stacklevel=1,
            )
            df_composed = as_conflict.drop_duplicates(pid=pid).reset_index(drop=True)

        elif resolve == "keep_all":
            warnings.warn(
                "Continuing, keeping all labels for duplicate records with inconsistent"
                " labels.",
                stacklevel=1,
            )
            df_composed = as_conflict.df

    else:
        df_composed = as_conflict.df

    # move included column to the end of dataframe
    included = df_composed.pop("included")
    df_composed = df_composed.assign(included=included)

    return df_composed


def _output_composition(final_df, output_file):
    # prepare collected labels to pass to the output file
    labels = [[index, row["included"]] for index, row in final_df.iterrows()]
    as_composed = ASReviewData(df=final_df)

    as_composed.to_file(output_file, labels=labels)

    print(f"Finished, exported composed dataset to {output_file}.")


def compose(
    output_file, rel, irr, lab, unl, pid="doi", order="riu", resolve="keep_one"
):
    # check whether all input has the same file extension
    _check_suffix([rel, irr, lab, unl], output_file)

    df_composition = create_composition(
        rel, irr, lab, unl, pid=pid, order=order, resolve=resolve
    )
    _output_composition(df_composition, output_file)


def _parse_arguments_compose():
    parser = argparse.ArgumentParser(prog="asreview data compose")
    parser.add_argument("output_path", type=str, help="The output file path.")
    parser.add_argument(
        "--relevant", "-r", type=str, help="A dataset with relevant records."
    )
    parser.add_argument(
        "--irrelevant", "-i", type=str, help="A dataset with irrelevant records."
    )
    parser.add_argument("--labeled", "-l", type=str, help="A labeled dataset.")
    parser.add_argument("--unlabeled", "-u", type=str, help="An unlabeled dataset.")
    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        type=_check_order_arg,
        default="riu",
        help="Hierarchy of labels in case of duplicates." "Default: riu.",
    )
    parser.add_argument(
        "--conflict_resolve",
        "-c",
        dest="conflict_resolve",
        type=_check_resolve_arg,
        default="keep_one",
        help="Method for dealing with " "conflicting labels.",
    )
    parser.add_argument(
        "--pid",
        type=str,
        default="doi",
        help="Persistent identifier used for deduplication. " "Default: doi.",
    )
    return parser



import argparse

from asreview.data import load_data
from asreview.entry_points import BaseEntryPoint

from asreviewcontrib.datatools import __version__
from asreviewcontrib.datatools.compose import _parse_arguments_compose
from asreviewcontrib.datatools.compose import compose
from asreviewcontrib.datatools.convert import _parse_arguments_convert
from asreviewcontrib.datatools.convert import convert
from asreviewcontrib.datatools.describe import _parse_arguments_describe
from asreviewcontrib.datatools.describe import describe
from asreviewcontrib.datatools.snowball import _parse_arguments_snowball
from asreviewcontrib.datatools.snowball import snowball
from asreviewcontrib.datatools.stack import _parse_arguments_vstack
from asreviewcontrib.datatools.stack import vstack

DATATOOLS = ["describe", "dedup", "convert", "compose", "vstack", "snowball"]


class DataEntryPoint(BaseEntryPoint):
    description = "Home of all data tools for ASReview."
    extension_name = "asreview-datatools"

    def __init__(self):
        from asreviewcontrib.datatools.__init__ import __version__

        super().__init__()

        self.version = __version__

    def execute(self, argv):
        if len(argv) > 1 and argv[0] in DATATOOLS:
            if argv[0] == "describe":
                args_describe_parser = _parse_arguments_describe()
                args_describe = vars(args_describe_parser.parse_args(argv[1:]))
                describe(**args_describe)
            if argv[0] == "convert":
                args_convert_parser = _parse_arguments_convert()
                args_convert = vars(args_convert_parser.parse_args(argv[1:]))
                convert(**args_convert)
            if argv[0] == "dedup":
                dedup_parser = argparse.ArgumentParser(prog="asreview data dedup")
                dedup_parser.add_argument(
                    "input_path", type=str, help="The file path of the dataset."
                )
                dedup_parser.add_argument(
                    "--output_path",
                    "-o",
                    default=None,
                    type=str,
                    help="The file path of the dataset.",
                )
                dedup_parser.add_argument(
                    "--pid",
                    default="doi",
                    type=str,
                    help="Persistent identifier used for deduplication. Default: doi.",
                )

                args_dedup = dedup_parser.parse_args(argv[1:])

                # read data in ASReview data object
                asdata = load_data(args_dedup.input_path)
                initial_length = len(asdata.df)

                if args_dedup.pid not in asdata.df.columns:
                    print(
                        f"Not using {args_dedup.pid} for deduplication"
                        "because there is no such data."
                    )

                # retrieve deduplicated ASReview data object
                asdata.drop_duplicates(pid=args_dedup.pid, inplace=True)

                # count duplicates
                n_dup = initial_length - len(asdata.df)

                if args_dedup.output_path:
                    asdata.to_file(args_dedup.output_path)
                    print(
                        f"Removed {n_dup} duplicates from dataset with"
                        f" {initial_length} records."
                    )
                else:
                    print(
                        f"Found {n_dup} duplicates in dataset with"
                        f" {initial_length} records."
                    )
            if argv[0] == "compose":
                args_compose_parser = _parse_arguments_compose()
                args_compose = args_compose_parser.parse_args(argv[1:])
                compose(
                    args_compose.output_path,
                    args_compose.relevant,
                    args_compose.irrelevant,
                    args_compose.labeled,
                    args_compose.unlabeled,
                    pid=args_compose.pid,
                    order=args_compose.hierarchy,
                    resolve=args_compose.conflict_resolve,
                )
            if argv[0] == "snowball":
                args_snowballing_parser = _parse_arguments_snowball()
                args_snowballing = vars(args_snowballing_parser.parse_args(argv[1:]))
                snowball(**args_snowballing)
            if argv[0] == "vstack":
                args_vstack_parser = _parse_arguments_vstack()
                args_vstack = args_vstack_parser.parse_args(argv[1:])
                vstack(args_vstack.output_path, args_vstack.datasets)

        # Print help message if subcommand not given or incorrect
        else:
            parser = argparse.ArgumentParser(
                prog="asreview data",
                formatter_class=argparse.RawTextHelpFormatter,
                description="Tools for data preprocessing for ASReview.",
            )
            parser.add_argument(
                "subcommand",
                nargs="?",
                default=None,
                help=f"The datatool to launch. Available commands:\n\n" f"{DATATOOLS}",
            )
            parser.add_argument(
                "-V",
                "--version",
                action="version",
                default=False,
                version=f"{self.extension_name}: {self.version}",
            )
            args, _ = parser.parse_known_args()

            print(args)
            # output the version
            if args.version:
                print(__version__)
                return

            parser.print_help()



import argparse
from pathlib import Path

import pandas as pd
from asreview import ASReviewData
from asreview.data.base import load_data


def _check_suffix(input_files, output_file):
    # Also raises ValueError on URLs that do not end with a file extension
    suffixes = [Path(item).suffix for item in input_files if item is not None]
    suffixes.append(Path(output_file).suffix)

    set_ris = {".txt", ".ris"}
    set_tabular = {".csv", ".tab", ".tsv", ".xlsx"}
    set_suffixes = set(suffixes)

    if len(set(suffixes)) > 1:
        if not (set_suffixes.issubset(set_ris) or set_suffixes.issubset(set_tabular)):
            raise ValueError(
                "• Several file types were given; All input files, as well as the"
                " output file should be of the same type. "
            )


def vstack(output_file, input_files):
    _check_suffix(input_files, output_file)

    list_dfs = [load_data(item).df for item in input_files]
    df_vstacked = pd.concat(list_dfs).reset_index(drop=True)
    as_vstacked = ASReviewData(df=df_vstacked)

    as_vstacked.to_file(output_file)


def _parse_arguments_vstack():
    parser = argparse.ArgumentParser(prog="asreview data vstack")
    parser.add_argument("output_path", type=str, help="The output file path.")
    parser.add_argument(
        "datasets",
        type=str,
        nargs="+",
        help="Any number of datasets to stack vertically.",
    )

    return parser



try:
    from asreviewcontrib.datatools._version import __version__
    from asreviewcontrib.datatools._version import __version_tuple__
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)



import argparse
import json

import asreview
from asreview.data import load_data
from asreview.data.statistics import n_duplicates
from asreview.data.statistics import n_irrelevant
from asreview.data.statistics import n_missing_abstract
from asreview.data.statistics import n_missing_title
from asreview.data.statistics import n_records
from asreview.data.statistics import n_relevant
from asreview.data.statistics import n_unlabeled

from asreviewcontrib.datatools import __version__


def describe(input_path, output_path=None):
    # read data in ASReview data object
    asdata = load_data(input_path)

    # based on https://google.github.io/styleguide/jsoncstyleguide.xml
    stats = {
        "asreviewVersion": asreview.__version__,
        "apiVersion": __version__,
        "data": {
            "items": [
                {
                    "id": "n_records",
                    "title": "Number of records",
                    "description": "The number of records in the dataset.",
                    "value": n_records(asdata),
                },
                {
                    "id": "n_relevant",
                    "title": "Number of relevant records",
                    "description": "The number of relevant records in the dataset.",
                    "value": n_relevant(asdata),
                },
                {
                    "id": "n_irrelevant",
                    "title": "Number of irrelevant records",
                    "description": "The number of irrelevant records in the dataset.",
                    "value": n_irrelevant(asdata),
                },
                {
                    "id": "n_unlabeled",
                    "title": "Number of unlabeled records",
                    "description": "The number of unlabeled records in the dataset.",
                    "value": n_unlabeled(asdata),
                },
                {
                    "id": "n_missing_title",
                    "title": "Number of records with missing title",
                    "description": (
                        "The number of records in the dataset with missing title."
                    ),
                    "value": n_missing_title(asdata)[0],
                },
                {
                    "id": "n_missing_abstract",
                    "title": "Number of records with missing abstract",
                    "description": (
                        "The number of records in the dataset with missing abstract."
                    ),
                    "value": n_missing_abstract(asdata)[0],
                },
                {
                    "id": "n_duplicates",
                    "title": "Number of duplicate records (basic algorithm)",
                    "description": (
                        "The number of duplicate records in the dataset based on"
                        " similar text."
                    ),
                    "value": n_duplicates(asdata),
                },
            ]
        },
    }  # noqa

    if output_path:
        with open(output_path, "w") as f:
            json.dump(stats, f, indent=2)

    print(json.dumps(stats, indent=2))


def _parse_arguments_describe():
    parser = argparse.ArgumentParser(prog="asreview data describe")
    parser.add_argument("input_path", type=str, help="The file path of the dataset.")
    parser.add_argument(
        "--output_path",
        "-o",
        default=None,
        type=str,
        help="The file path of the dataset.",
    )

    return parser



from pathlib import Path

import pandas as pd

from asreviewcontrib.datatools.snowball import backward_snowballing
from asreviewcontrib.datatools.snowball import forward_snowballing
from asreviewcontrib.datatools.snowball import openalex_from_doi
from asreviewcontrib.datatools.snowball import snowball

INPUT_DIR = Path(__file__).parent / "demo_data"


def test_openalex_from_doi():
    dois = [
        "https://doi.org/10.1042/cs20220150",
        "https://doi.org/10.1042/bst20220734",
        "not_a_doi",
    ]

    assert openalex_from_doi(dois) == {
        "10.1042/cs20220150": "https://openalex.org/W4386305682",
        "10.1042/bst20220734": "https://openalex.org/W4312006214",
        "not_a_doi": None,
    }


def test_backward_snowballing():
    identifiers = [
        "https://openalex.org/W4281483266",
        "https://openalex.org/W2008620264",
    ]

    backwards_citations = backward_snowballing(identifiers)

    assert "https://openalex.org/W1864285629" in [
        field_dict["id"] for field_dict in backwards_citations[identifiers[0]]
    ]
    assert "https://openalex.org/W950821216" in [
        field_dict["id"] for field_dict in backwards_citations[identifiers[1]]
    ]


def test_forward_snowballing():
    identifiers = [
        "https://openalex.org/W4281483266",
        "https://openalex.org/W2008620264",
    ]

    forwards_citations = forward_snowballing(identifiers)

    assert "https://openalex.org/W4386305682" in [
        field_dict["id"] for field_dict in forwards_citations[identifiers[0]]
    ]
    assert "https://openalex.org/W2124637492" in [
        field_dict["id"] for field_dict in forwards_citations[identifiers[1]]
    ]


def test_openalex_id_forward(tmpdir):
    out_fp = Path(tmpdir, "forward_all.csv")
    snowball(
        input_path=INPUT_DIR / "snowballing_openalex.csv",
        output_path=out_fp,
        forward=True,
        backward=False,
        use_all=False,
    )
    df = pd.read_csv(out_fp)
    assert len(df) >= 23

    all_out_fp = Path(tmpdir, "forward_all.csv")
    snowball(
        input_path=INPUT_DIR / "snowballing_openalex.csv",
        output_path=all_out_fp,
        forward=True,
        backward=False,
        use_all=True,
    )
    df_all = pd.read_csv(all_out_fp)
    assert len(df_all) >= 387


def test_openalex_id_backward(tmpdir):
    out_fp = Path(tmpdir, "forward_all.csv")
    snowball(
        input_path=INPUT_DIR / "snowballing_openalex.csv",
        output_path=out_fp,
        forward=False,
        backward=True,
        use_all=False,
    )
    df = pd.read_csv(out_fp)
    assert len(df) == 31

    all_out_fp = Path(tmpdir, "backward_all.csv")
    snowball(
        input_path=INPUT_DIR / "snowballing_openalex.csv",
        output_path=all_out_fp,
        forward=False,
        backward=True,
        use_all=True,
    )
    df_all = pd.read_csv(all_out_fp)
    assert len(df_all) == 117


def test_snowballing_from_doi(tmpdir):
    out_fp = Path(tmpdir, "doi_all.csv")
    snowball(
        input_path=INPUT_DIR / "snowballing_doi.csv",
        output_path=out_fp,
        forward=False,
        backward=True,
        use_all=True,
    )
    df = pd.read_csv(out_fp)
    assert len(df) == 117



from pathlib import Path

import pytest

from asreviewcontrib.datatools.compose import _check_order_arg
from asreviewcontrib.datatools.compose import _check_resolve_arg
from asreviewcontrib.datatools.compose import _check_suffix
from asreviewcontrib.datatools.compose import create_composition

parent_dir = Path(__file__).parent
file_1 = Path(parent_dir, "demo_data", "dataset_1.ris")
file_2 = Path(parent_dir, "demo_data", "dataset_2.ris")

# labeling action on input paths in list = [relevant, irrelevant, labeled, unlabeled]
input_files_1 = [
    file_1,
    file_1,
    file_1,
    file_1,
]

input_files_2 = [None, None, file_1, file_2]


# test whether input and output suffixes are compatible
def test_suffixes():
    with pytest.raises(ValueError):
        _check_suffix(input_files_1, "conflicting_suffix.csv")


# test whether wrong input hierarchy/order raises error
def test_input_hierarchy():
    with pytest.raises(ValueError):
        _check_order_arg("abc")
    with pytest.raises(ValueError):
        _check_order_arg("riur")


# test whether wrong input conflict resolve raises error
def test_input_resolve():
    with pytest.raises(ValueError):
        _check_resolve_arg("fly")


def test_label_prioritization():
    # input identical datasets and overwrite everything with the relevant labels
    df_1 = create_composition(*input_files_1, order="riu")
    assert df_1["included"].value_counts()[1] == len(df_1)

    # input identical datasets and overwrite everything with the irrelevant labels
    df_2 = create_composition(*input_files_1, order="iru")
    assert df_2["included"].value_counts()[0] == len(df_2)

    # input identical datasets and overwrite everything as unlabeled
    df_3 = create_composition(*input_files_1, order="uri")
    assert df_3["included"].value_counts()[-1] == len(df_3)

    # input different datasets with some identical records, combining as labeled and
    # unlabeled data
    df_4 = create_composition(*input_files_2, order="riu")
    df_4_counts = df_4["included"].value_counts()
    assert df_4_counts[-1] == 7 and df_4_counts[0] == 3 and df_4_counts[1] == 1



import argparse

from asreview.data import ASReviewData


def convert(input_path, output_path):
    # read data in ASReview data object
    asdata = ASReviewData.from_file(input_path)

    asdata.to_file(output_path)


def _parse_arguments_convert():
    parser = argparse.ArgumentParser(prog="asreview data convert")
    parser.add_argument("input_path", type=str, help="The file path of the dataset.")
    parser.add_argument("output_path", type=str, help="The file path of the dataset.")
    return parser



from pathlib import Path

from asreview.data import ASReviewData

from asreviewcontrib.datatools.stack import vstack

test_dir = Path(__file__).parent
file_1 = Path(test_dir, "demo_data", "dataset_1.ris")
file_2 = Path(test_dir, "demo_data", "dataset_2.ris")


def test_stack(tmpdir):
    output_path = Path(tmpdir, "test_output.ris")
    vstack(output_path, [file_1, file_2])
    as_test = ASReviewData.from_file(output_path)

    assert len(as_test.df) == 14
    assert as_test.df["included"].value_counts()[-1] == 9
    assert as_test.df["included"].value_counts()[0] == 3
    assert as_test.df["included"].value_counts()[1] == 2



from __future__ import annotations

import argparse
from pathlib import Path

import pandas as pd
import pyalex
from asreview import ASReviewData
from asreview import load_data

# Maximum number of statements joined by a logical OR in a call to OpenAlex.
OPENALEX_MAX_OR_LENGTH = 100
OPENALEX_MAX_PAGE_LENGTH = 200
OPENALEX_PREFIX = "https://openalex.org/"
DOI_PREFIX = "https://doi.org/"

# OpenAlex data fields to retrieve.
USED_FIELDS = [
    "id",
    "doi",
    "title",
    "abstract_inverted_index",
    "referenced_works",
    "publication_date",
]


def forward_snowballing(identifiers: list[str]) -> dict[str, list[dict]]:
    """Get all works citing a work with the OpenAlex identifier from the list.

    Parameters
    ----------
    identifiers : list[str]
        List of OpenAlex identifiers.

    Returns
    -------
    dict[str, list[dict]]
        Dictionary of the form
            `{input OpenAlex identifier : list of OpenAlex works}`
        where each work in the list references the work with the input identifier and
        it is a dictionary of the form `{field_name : field_value}`.
    """
    citing_works = {}
    for idx, openalex_id in enumerate(identifiers):
        print(f"{idx}. Getting works citing {openalex_id}")
        pager = (
            pyalex.Works()
            .filter(cites=openalex_id)
            .select(USED_FIELDS)
            .paginate(per_page=OPENALEX_MAX_PAGE_LENGTH, n_max=None)
        )
        citing_works[openalex_id] = []
        for page in pager:
            citing_works[openalex_id] += [
                {
                    key: work[key]
                    for key in [
                        col if col != "abstract_inverted_index" else "abstract"
                        for col in USED_FIELDS
                    ]
                }
                for work in page
            ]
    return citing_works


def backward_snowballing(identifiers: list[str]) -> dict[str, list[dict]]:
    """Get all works cited by a work with the OpenAlex identifier from the list.

    Parameters
    ----------
    identifiers : list[str]
        List of OpenAlex identifiers.

    Returns
    -------
    dict[str, list[dict]]
        Dictionary of the form
            `{input OpenAlex identifier : list of OpenAlex works}`
        where each work in the list is referenced by the work with the input identifier
        and it is a dictionary of the form `{field_name : field_value}`.
    """
    # Get the referenced works.
    referenced_works = {}
    page_length = min(OPENALEX_MAX_OR_LENGTH, OPENALEX_MAX_PAGE_LENGTH)

    for i in range(0, len(identifiers), page_length):
        print(f"Getting works citing records {i}-{i+page_length}")
        # We need to remove the prefix here because otherwise the URL is too long.
        fltr = "|".join(
            identifier.removeprefix(OPENALEX_PREFIX)
            for identifier in identifiers[i : i + page_length]
        )
        for work in (
            pyalex.Works()
            .filter(openalex=fltr)
            .select("id,referenced_works")
            .get(per_page=page_length)
        ):
            referenced_works[work["id"]] = work["referenced_works"]

    # Get the fields for the referenced works.
    all_identifiers = []
    for reference_list in referenced_works.values():
        all_identifiers += reference_list
    all_identifiers = list(set(all_identifiers))
    print(f"Found {len(all_identifiers)} records")

    all_referenced_works = {}
    for i in range(0, len(all_identifiers), page_length):
        # We need to remove the prefix here because otherwise the URL is too long.
        fltr = "|".join(
            identifier.removeprefix(OPENALEX_PREFIX)
            for identifier in all_identifiers[i : i + page_length]
        )
        for work in (
            pyalex.Works()
            .filter(openalex=fltr)
            .select(USED_FIELDS)
            .get(per_page=page_length)
        ):
            all_referenced_works[work["id"]] = {
                key: work[key]
                for key in [
                    col if col != "abstract_inverted_index" else "abstract"
                    for col in USED_FIELDS
                ]
            }

    # Connect the referenced works back to the input works.
    output = {}
    for identifier, ref_id_list in referenced_works.items():
        # We need the last check if 'ref_id' is in 'all_referenced_works': If a work
        # references an ID that redirects to another ID, it won't be present here.
        # Example: https://openalex.org/W2015370450 has in the references the identifier
        # https://openalex.org/W2008744335, but this redirects to
        # https://openalex.org/W4233569835
        output[identifier] = [
            all_referenced_works[ref_id]
            for ref_id in ref_id_list
            if ref_id in all_referenced_works
        ]
    return output


def openalex_from_doi(dois: list[str]) -> dict[str, str]:
    """Get the OpenAlex identifiers corresponding to a list of DOIs.

    Parameters
    ----------
    dois : list[str]
        List of DOIs.

    Returns
    -------
    dict[str, str]
        Dictionary {doi: openalex_id}. If there was no OpenAlex identifier found for a
        DOI, the corresponding value will be None.
    """
    page_length = min(OPENALEX_MAX_OR_LENGTH, OPENALEX_MAX_PAGE_LENGTH)
    id_mapping = {doi.removeprefix(DOI_PREFIX): None for doi in dois}
    for i in range(0, len(dois), page_length):
        fltr = "|".join(dois[i : i + page_length])
        for work in (
            pyalex.Works()
            .filter(doi=fltr)
            .select(["id", "doi"])
            .get(per_page=page_length)
        ):
            id_mapping[work["doi"].removeprefix(DOI_PREFIX)] = work["id"]
    return id_mapping


def snowball(
    input_path: Path,
    output_path: Path,
    forward: bool,
    backward: bool,
    use_all: bool = False,
    email: str = None,
) -> None:
    """Perform snowballing on an ASReview dataset.

    Parameters
    ----------
    input_path : Path
        Location of the input ASReview dataset.
    output_path : Path
        Location where to save the output dataset.
    forward : bool
        Perform forward snowballing. At least one of `forward` or `backward` should be
        True.
    backward : bool
        Perform backward snowballing. At least one of `forward` or `backward` should be
        True.
    use_all : bool, optional
        Perform snowballing on all records in the dataset or only the included
        records, by default False
    email : str, optional
        Email address to send along with request to OpenAlex, by default None

    Raises
    ------
    ValueError
        If `forward` and `backward` are both False.
    ValueError
        If the dataset contains no column name `openalex_id` and no column names `doi`.
    """
    if not (forward or backward):
        raise ValueError("At least one of 'forward' or 'backward' should be True.")

    data = load_data(input_path)
    if use_all or (data.included is None):
        data = data.df
    else:
        data = data.df.loc[data.included.astype(bool)]

    # Add OpenAlex identifiers if not available.
    if "openalex_id" not in data.columns:
        if "doi" not in data.columns:
            raise ValueError(
                "Dataset should contain a column 'openalex_id' containing OpenAlex"
                " identifiers or a column 'doi' containing DOIs."
            )
        id_mapping = openalex_from_doi(data.doi.dropna().to_list())
        n_openalex_ids = len(
            [
                openalex_id
                for openalex_id in id_mapping.values()
                if openalex_id is not None
            ]
        )
        print(
            f"Found OpenAlex identifiers for {n_openalex_ids} out of {len(data)}"
            " records. Performing snowballing for those records."
        )
        data["openalex_id"] = None
        data.loc[data.doi.notna(), "openalex_id"] = (
            data.loc[data.doi.notna(), "doi"]
            .str.removeprefix(DOI_PREFIX)
            .apply(lambda doi: id_mapping[doi])
        )

    identifiers = data["openalex_id"].dropna().to_list()

    if email is not None:
        pyalex.config.email = email

    if forward:
        print("Starting forward snowballing")
        forward_data = forward_snowballing(identifiers)
    else:
        forward_data = {}
    if backward:
        print("Starting backward snowballing")
        backward_data = backward_snowballing(identifiers)
    else:
        backward_data = {}

    all_works = []
    for works_list in forward_data.values():
        all_works += works_list
    for works_list in backward_data.values():
        all_works += works_list

    output_data = pd.DataFrame(all_works)
    output_data.drop_duplicates(subset=["id"], inplace=True)
    output_data.rename({"id": "openalex_id"}, axis=1, inplace=True)
    output_data = ASReviewData(output_data)
    output_data.to_file(output_path)
    print("Saved dataset")


def _parse_arguments_snowball():
    parser = argparse.ArgumentParser(prog="asreview data snowballing")
    parser.add_argument(
        "input_path", type=str, help="The file path of the input dataset."
    )
    parser.add_argument(
        "output_path", type=str, help="The file path of the output dataset."
    )
    parser.add_argument(
        "--forward", "-f", action="store_true", help="Do forward snowballing."
    )
    parser.add_argument(
        "--backward", "-b", action="store_true", help="Do backward snowballing."
    )
    parser.add_argument(
        "--all",
        "-a",
        action="store_true",
        dest="use_all",
        help=(
            "Do snowballing on all records in the dataset, not just the included ones."
        ),
    )
    parser.add_argument(
        "--email",
        "-e",
        type=str,
        required=False,
        help=(
            "Email address to send along with requests to OpenAlex. This will make"
            " requests faster. See also "
            "https://docs.openalex.org/how-to-use-the-api/rate-limits-and-authentication#the-polite-pool"
        ),
    )
    return parser
